/* Copyright 2007, 2012 Kevin Ryde.

   This file is part of x2gpm.

   x2gpm is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   x2gpm is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with X2gpm; see the file COPYING.  Failing that, goto
   <http://www.gnu.org/licenses/>. */

#define _GNU_SOURCE
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <iconv.h>
#include <limits.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>     /* for NULL */
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <utmp.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/poll.h>

#undef WUNTRACED  /* avoid redefines from <linux/wait.h> */
#undef WNOHANG
#undef WSTOPPED
#undef WEXITED
#undef WCONTINUED
#include <linux/keyboard.h>  /* for KG_CTRL */


/* #define TRACE(x) */
/* int option_all = 0; */
/* #include <ctype.h> */
/* #include <errno.h> */
/* #include <string.h> */
/* #include <malloc.h> */


iconv_t
open_iconv (iconv_t *cvp, const char *to, const char *from)
{
  if (*cvp == (iconv_t) -1)
    {
      *cvp = iconv_open (to, from);
      if (*cvp == (iconv_t) -1)
        {
          syslog (LOG_ERR, "cannot iconv %s to %s, pasting raw", from, to);
          beep ();
          return (iconv_t) -1;
        }
    }
  return *cvp;
}

int
console_is_utf8 (int fd, const char *filename)
{
  static int warned = 0;
  long mode;
  if (ioctl (fd, KDGKBMODE, &mode) != 0)
    {
      /* error will be something fishy, maybe an ancient kernel? */
      if (! warned)
        {
          syslog (LOG_ERR, "ioctl KDGKBMODE %s error, assuming 8-bit: %s",
                  filename, strerror (errno));
          warned = 1;
        }
      return 0;
    }
  return (mode == K_UNICODE);
}

int
tiocsti_raw (int fd, const char *filename, const char *buf, size_t len)
{
  size_t i;
  for (i = 0; i < len; i++)
    {
      if (ioctl (fd, TIOCSTI, &buf[i]) != 0)
        {
          syslog (LOG_ERR, "ioctl TIOCSTI %s error: %s",
                  filename, strerror (errno));
          beep ();
          return 0;
        }
    }
  return 1;
}

void
tiocsti_cooked (const char *s)
{
  static int  fd = -1;
  char  *filename = (option_all ? "/dev/console" : "/dev/tty");
  iconv_t cv = (iconv_t) -1;

  TRACE (printf ("TIOCSTI: \"%s\"\n", s));

  if (fd == -1)
    {
      fd = open (filename, O_WRONLY);
      if (fd == -1)
        {
          syslog (LOG_ERR, "cannot open %s: %s", filename, strerror (errno));
          beep ();
          return;
        }
    }

  if (console_is_utf8 (fd, filename))
    {
      static iconv_t STRING_to_utf8 = (iconv_t) -1;
      cv = open_iconv (&STRING_to_utf8, "latin-1", "utf-8");
    }

  if (cv == (iconv_t) -1)
    {
      tiocsti_raw (fd, filename, s, strlen (s));
    }
  else
    {
      char  out_buf[256];
      size_t out_alloc = sizeof (out_buf);
      char   *in_ptr, *out_ptr;
      size_t in_len, out_left;

      /* converted paste */
      if (iconv (cv, NULL, NULL, NULL, NULL) == -1)
        {
          syslog (LOG_ERR, "iconv error restting: %s", strerror (errno));
          beep ();
          goto done;
        }

      in_ptr = (char *) s;
      in_len = strlen (s);

      while (in_len != 0)
        {
          out_ptr = out_buf;
          out_left = out_alloc;
          if (iconv (cv, &in_ptr, &in_len, &out_ptr, &out_left) == -1)
            {
              syslog (LOG_ERR, "iconv error: %s", strerror (errno));
              beep ();
              goto done;
            }
          if (! tiocsti_raw (fd, filename, out_buf, out_alloc - out_left))
            goto done;
        }
    }

 done:
  if (option_all)
    {
      if (close (fd) != 0)
        {
          syslog (LOG_ERR, "error closing %s: %s", filename, strerror (errno));
          beep ();
        }
      fd = -1;
    }
}
